from __future__ import annotations

import abc
import json
from typing import List, Optional, Union, Dict

from bigeye_sdk.client.enum import Method
from bigeye_sdk.functions.metric_functions import is_same_metric, is_same_column_metric
from bigeye_sdk.functions.urlfuncts import encode_url_params
from bigeye_sdk.generated.com.bigeye.models.generated import TableList, ColumnMetricProfileList, \
    AutometricAcceptRequest, \
    AutometricAcceptResponse, GetMetricInfoListRequest, MetricInfoList, SearchMetricConfigurationRequest, \
    BatchGetMetricResponse, MetricCreationState, ThreeLeggedBoolean, MetricSortField, SortDirection, \
    MetricConfiguration, \
    MetricBackfillResponse, MetricBackfillRequest, GetCollectionsResponse, GetCollectionResponse, Collection, \
    EditCollectionResponse, EditCollectionRequest, \
    GetDeltaApplicableMetricTypesResponse, BatchRunMetricsRequest, BatchRunMetricsResponse, \
    GetSourceListResponse, Empty, Source, CreateSourceRequest, BatchMetricConfigRequest, \
    BatchMetricConfigResponse, Table, ComparisonTableInfo, GetComparisonTableInfosResponse, RunComparisonTableResponse, \
    GetComparisonTableInfosRequest, GetIssuesRequest, Issue, GetIssuesResponse, UpdateIssueRequest, \
    IssueStatusUpdate, IssueStatus, MetricRunLabel, User, UpdateIssueResponse, \
    GetTableComparisonMetricGroupInfoRequest, GetTableComparisonMetricGroupInfoResponse, ComparisonMetricGroup, \
    GetPreviewResponse, NamedSchedule, CreateNamedScheduleRequest, \
    GetNamedSchedulesRequest, NamedScheduleSortField, GetNamedSchedulesResponse, \
    SchemaList, SchemaSearchRequest, DataNodeType, CreateDataNodeRequest, DataNode, \
    CreateLineageRelationshipRequest, LineageRelationship, CreateCollectionResponse, TimeRange, NotificationChannel, \
    CreateMetricTemplateRequest, MetricTemplate, MetricTemplateParametersFieldEntry, Warehouse, \
    CreateMetricTemplateResponse, VirtualTableRequest, VirtualTable, MetricTemplateParameterType, FieldType, \
    GetMetricTemplateListRequest, GetMetricTemplateListResponse, MetricTemplateSortField, UpdateVirtualTableRequest, \
    SchemaChangeSortField, GetSchemaChangesRequest, GetSchemaChangesResponse, SchemaChange, IntegrationsResponse, \
    TableauWorkbook, Integration, GetIntegrationEntitiesResponse, ConfigValue, GetConfigListResponse, WorkflowResponse, \
    WorkflowStatusResponse, GetLineageRelationshipsForNodeResponse

# create logger
from bigeye_sdk.log import get_logger
from bigeye_sdk.model.protobuf_enum_facade import SimpleFieldType, SimpleMetricTemplateParameterType

log = get_logger(__file__)


class GeneratedDatawatchClient(abc.ABC):
    """TODO: In future, should be generated and only contain methods generated from protobuf. :)"""

    def _call_datawatch(self, method: Method, url, body: str = None, params: dict = None):
        url = url.replace('//', '/')
        return self._call_datawatch_impl(method=method, url=url, body=body, params=params)

    @abc.abstractmethod
    def _call_datawatch_impl(self, method: Method, url, body: str = None, params: dict = None):
        """Each implementation must override this."""
        pass

    def get_dataset(self, warehouse_id: int, schema_name: str) -> List[dict]:
        """
        Provides advanced dataset details for a schema but returns a list of dictionaries instead of an list of Table
        objects.

        Args:
            warehouse_id:
            schema_name:

        Returns: list of dictionaries containing advanced details about all datasets in a schema.

        """
        url = f"/dataset/tables/{warehouse_id}/{schema_name}"

        return self._call_datawatch(Method.GET, url)

    def get_tables(self,
                   *,
                   warehouse_id: List[int] = [],
                   schema: List[str] = [],
                   table_name: List[str] = [],
                   ids: List[int] = [],
                   schema_id: List[int] = []) -> TableList:
        url = f"/api/v1/tables{encode_url_params(locals(), remove_keys=['api_conf'])}"
        log.info('Getting warehouse tables.')
        log.info(url)
        response = self._call_datawatch(Method.GET, url)
        tables = TableList().from_dict(response)
        return tables

    def get_table_ids(self,
                      warehouse_id: List[int] = [],
                      schemas: List[str] = [],
                      table_name: List[str] = [],
                      ids: List[int] = [],
                      schema_id: List[int] = []) -> List[int]:
        return [t.id for t in self.get_tables(warehouse_id=warehouse_id, schema=schemas, table_name=table_name,
                                              ids=ids, schema_id=schema_id).tables]

    def rebuild(self, warehouse_id: int, schema_name: str = None):
        """
        TODO: switch to returning an object from the protobuf.

        Args:
            warehouse_id:
            schema_name:

        Returns: list of metrics.

        """
        url = f'/dataset/rebuild/{warehouse_id}'

        if schema_name is not None:
            url = url + f'/{schema_name}'

        return self._call_datawatch(Method.GET, url)

    def get_table_profile(self, table_id: int) -> ColumnMetricProfileList:
        url = f'/api/v1/tables/profile/{table_id}'
        return ColumnMetricProfileList().from_dict(self._call_datawatch(Method.GET, url))

    def batch_delete_metrics(self, metric_ids: List[int]):
        body = json.dumps({"metricIds": metric_ids})
        log.info(f'Deleting metrics: {metric_ids}')
        url = f'/statistics'
        return self._call_datawatch(Method.DELETE, url, body)

    def accept_autometrics(
            self, *, metric_ids: List[int] = [], lookback_days: int = 7
    ) -> AutometricAcceptResponse:
        """Batch accept autometrics"""

        request = AutometricAcceptRequest()
        request.metric_ids = metric_ids
        request.lookback_days = lookback_days

        url = f'/api/v1/metrics/autometrics/accept'
        body = request.to_json()
        log.info(f'Deploying autometrics: {url} -- {body}')
        r = self._call_datawatch(Method.POST, url, body)
        return AutometricAcceptResponse().from_dict(r)

    def get_metric_configuration(
            self, *, metric_id: int = 0
    ) -> MetricConfiguration:
        """Get metric configuration"""

        url = f'/api/v1/metrics/{metric_id}'

        r = self._call_datawatch(Method.GET, url)
        return MetricConfiguration().from_dict(r)

    def search_metric_configuration(
            self,
            *,
            ids: List[int] = [],
            warehouse_ids: List[int] = [],
            table_ids: List[int] = [],
            table_name: str = "",
            status: str = "",
            muted: bool = False,
    ) -> List[MetricConfiguration]:
        """Search metric configurations"""

        request = SearchMetricConfigurationRequest()
        request.ids = ids
        request.warehouse_ids = warehouse_ids
        request.table_ids = table_ids
        request.table_name = table_name
        request.status = status
        request.muted = muted

        url = f'/api/v1/metrics{encode_url_params(d=request.to_dict())}'

        response = self._call_datawatch(Method.GET, url=url)

        return [MetricConfiguration().from_dict(m) for m in BatchGetMetricResponse(metrics=response).metrics]

    def get_metric_info_batch_post(
            self,
            *,
            metric_ids: List[int] = [],
            warehouse_ids: List[int] = [],
            table_ids: List[int] = [],
            table_name: str = "",
            status: str = "",  # healthy, alerting, unknown
            metric_creation_states: List[MetricCreationState] = [],
            muted: ThreeLeggedBoolean = 0,
            page_size: int = 0,
            page_cursor: str = "",
            sort_field: MetricSortField = 0,
            schema_name: str = "",
            column_ids: List[int] = [],
            search: str = "",
            sort_direction: SortDirection = 0,
    ) -> MetricInfoList:
        """
        Get metric information as batch.
        Args:
            metric_ids: list of metric ids
            warehouse_ids: list of source ids
            table_ids: list of table ids
            table_name: a single table name
            status: healthy, alerting, unknown
            metric_creation_states: a list of metric creation states (enum)
            muted: three-legged true, false, unspecified
            page_size: int
            page_cursor: unique id of the cursor
            sort_field: sort by value MetricSortField (enum)
            schema_name: a single schema name
            column_ids: a list of column ids
            search: a raw search of metric names.
            sort_direction: three-legged ascending, descending, unspecified

        Returns: a list of metric info fitting the argument criteria.

        """

        request = GetMetricInfoListRequest()
        request.metric_ids = metric_ids
        request.warehouse_ids = warehouse_ids
        request.table_ids = table_ids
        request.table_name = table_name
        request.status = status
        request.metric_creation_states = metric_creation_states
        request.muted = muted
        request.page_size = page_size
        request.page_cursor = page_cursor
        request.sort_field = sort_field
        request.schema_name = schema_name
        request.column_ids = column_ids
        request.search = search
        request.sort_direction = sort_direction

        url = '/api/v1/metrics/info'

        if len(request.to_dict()) != 0:
            response = self._call_datawatch(Method.POST, url=url, body=request.to_json())
        else:
            response = self._call_datawatch(Method.GET, url=url)

        mil_current = MetricInfoList().from_dict(response)
        mil_return = MetricInfoList()
        mil_return.metrics.extend(mil_current.metrics)

        while mil_current.pagination_info.next_cursor:
            request.page_cursor = mil_current.pagination_info.next_cursor
            response = self._call_datawatch(Method.POST, url=url, body=request.to_json())
            mil_current = MetricInfoList().from_dict(response)
            mil_return.metrics.extend(mil_current.metrics)

        return mil_return

    def get_existing_metric(self,
                            warehouse_id: int, table: Table, column_name: str, user_defined_name: str,
                            metric_name: str, group_by: List[str], filters: List[str]):
        """
        Get an existing metric by name and group_by.
        """
        metrics: List[MetricConfiguration] = self.search_metric_configuration(warehouse_ids=[warehouse_id],
                                                                              table_ids=[table.id])

        for m in metrics:
            if is_same_metric(m, metric_name, user_defined_name, group_by, filters) \
                    and is_same_column_metric(m, column_name):
                return m
        return None

    def backfill_metric(
            self,
            *,
            metric_ids: List[int] = [],
            backfill_range: Optional[TimeRange] = None,
            delete_history: Optional[bool] = None
    ) -> MetricBackfillResponse:
        """
        Runs metrics for past data and returns the API response.  Destructive.
        :param metric_ids: list of metric ids to run.
        :param backfill_range: time range for metrics to be run.
        :param delete_history: whether to delete metric run history with backfill.
        :return: MetricBackfillResponse object.
        """

        request = MetricBackfillRequest()
        request.metric_ids = metric_ids
        if backfill_range is not None:
            request.backfill_range = backfill_range
        if delete_history is not None:
            request.delete_history = delete_history

        url = "/api/v1/metrics/backfill"

        response = self._call_datawatch(Method.POST, url=url, body=request.to_json())

        return MetricBackfillResponse().from_dict(response)

    def get_collections(self) -> GetCollectionsResponse:
        url = "/api/v1/collections/"
        log.info(url)
        response = self._call_datawatch(Method.GET, url=url)
        return GetCollectionsResponse().from_dict(response)

    def get_collection(self, collection_id: int) -> GetCollectionResponse:
        url = f"/api/v1/collections/{collection_id}"
        log.info(url)
        response = self._call_datawatch(Method.GET, url=url)
        return GetCollectionResponse().from_dict(response)

    def create_collection(
            self,
            *,
            collection_name: str = "",
            description: str = "",
            metric_ids: List[int] = [],
            notification_channels: List[NotificationChannel] = [],
            muted_until_timestamp: int = 0,
    ) -> CreateCollectionResponse:
        """Create collection"""

        url = f"/api/v1/collections"

        request = EditCollectionRequest()
        request.collection_name = collection_name
        request.description = description
        request.metric_ids = metric_ids
        if notification_channels is not None:
            request.notification_channels = notification_channels
        request.muted_until_timestamp = muted_until_timestamp

        log.info(f'Query: {url}; Body: {request.to_json()}')

        response = self._call_datawatch(Method.POST, url=url, body=request.to_json())

        return CreateCollectionResponse().from_dict(response)

    def create_collection_dep(self, collection: Collection) -> EditCollectionResponse:
        url = f"/api/v1/collections"

        request: EditCollectionRequest = EditCollectionRequest()
        request.collection_name = collection.name
        request.description = collection.description
        request.metric_ids = collection.metric_ids
        request.notification_channels = collection.notification_channels
        request.muted_until_timestamp = collection.muted_until_timestamp

        log.info(f'Query: {url}; Body: {request.to_json()}')

        response = self._call_datawatch(Method.POST, url=url, body=request.to_json())

        return EditCollectionResponse().from_dict(response)

    def update_collection(self, collection: Collection) -> EditCollectionResponse:

        request: EditCollectionRequest = EditCollectionRequest()
        request.collection_name = collection.name
        request.description = collection.description
        request.metric_ids = collection.metric_ids
        request.notification_channels = collection.notification_channels
        request.muted_until_timestamp = collection.muted_until_timestamp

        url = f"/api/v1/collections/{collection.id}"

        log.info(f'Query: {url}; Body: {request.to_json()}')

        response = self._call_datawatch(Method.PUT, url=url, body=request.to_json())

        return EditCollectionResponse().from_dict(response)

    def upsert_metric_to_collection(self, collection_id: int,
                                    add_metric_ids: Union[int, List[int]]) -> EditCollectionResponse:

        if type(add_metric_ids) == int:
            mids = list(add_metric_ids)
        else:
            mids = add_metric_ids

        collection: Collection = self.get_collection(collection_id=collection_id).collection

        for mid in mids:
            collection.metric_ids.append(mid)

        return self.update_collection(collection=collection)

    def delete_collection(self, collection_id: int):
        url = f"/api/v1/collections/{collection_id}"

        self._call_datawatch(Method.DELETE, url=url)

    def delete_delta(self, *, comparison_table_id: int = 0):
        """Delete a delta"""

        url = f"/api/v1/metrics/comparisons/tables/{comparison_table_id}"

        self._call_datawatch(Method.DELETE, url)

    def get_delta_applicable_metric_types(
            self, *, table_id: int = 0
    ) -> GetDeltaApplicableMetricTypesResponse:
        """
        Get list of metrics applicable for deltas
        Args:
            table_id: source table id

        Returns: list of metrics applicable for deltas.

        """

        url = f"/api/v1/tables/{table_id}/delta-applicable-metric-types"

        response = self._call_datawatch(Method.GET, url)

        return GetDeltaApplicableMetricTypesResponse().from_dict(response)

    def run_a_delta(self, *, delta_id: int) -> ComparisonTableInfo:
        """

        Args:
            delta_id: Required.  The ID of the delta to run

        Returns:  ComparisonTableInfo

        """

        url = f"/api/v1/metrics/comparisons/tables/run/{delta_id}"
        response = self._call_datawatch(Method.GET, url)

        return RunComparisonTableResponse().from_dict(response).comparison_table_info

    def get_delta_information(self, *, delta_ids: List[int],
                              exclude_comparison_metrics: bool = False) -> List[ComparisonTableInfo]:
        """

            Args:
                delta_ids: Required.  The delta ID or IDs.
                exclude_comparison_metrics: Optional. Whether to include the list of ComparisonMetricInfos

            Returns:  List[ComparisonTableInfo]

        """

        url = "/api/v1/metrics/comparisons/tables/info"
        request = GetComparisonTableInfosRequest(delta_ids, exclude_comparison_metrics)
        response = self._call_datawatch(Method.POST, url, request.to_json())

        return GetComparisonTableInfosResponse().from_dict(response).comparison_table_infos

    def get_delta_groups_information(self, *, comparison_metric_id: int) -> List[ComparisonMetricGroup]:
        """

            Args:
                comparison_metric_id: Required.  The ID for a delta with grouped by columns.

            Returns:  List[ComparisonMetricGroup]

        """

        url = f"/api/v1/metrics/comparisons/metrics/{comparison_metric_id}/groups/info"
        request = GetTableComparisonMetricGroupInfoRequest(comparison_metric_id)
        response = self._call_datawatch(Method.GET, url, request.to_json())

        return GetTableComparisonMetricGroupInfoResponse().from_dict(response).group_state

    def run_metric_batch(
            self, *, metric_ids: List[int] = []
    ) -> BatchRunMetricsResponse:
        """Batch run metrics"""

        request = BatchRunMetricsRequest()
        request.metric_ids = metric_ids

        url = '/api/v1/metrics/run/batch'

        response = self._call_datawatch(Method.POST, url, request.to_json())

        return BatchRunMetricsResponse().from_dict(response)

    def get_sources(self) -> GetSourceListResponse:
        """Get sources"""
        url = "/api/v1/sources/fetch"
        request = Empty()

        response = self._call_datawatch(Method.POST, url, request.to_json())

        return GetSourceListResponse().from_dict(response)

    def get_schemas(self, *, warehouse_id: List[int] = []) -> SchemaList:
        request = SchemaSearchRequest()
        request.warehouse_id = warehouse_id

        url = f'/api/v1/schemas{encode_url_params(d=request.to_dict())}'

        response = self._call_datawatch(Method.GET, url)

        return SchemaList().from_dict(response)

    def edit_metric(self, metric_configuration: MetricConfiguration = None) -> BatchMetricConfigResponse:
        """
        TODO: Only supporting 1 metric because of weird defaults and required fields.
        Args:
            metric_configuration:

        Returns:

        """
        url = "/api/v1/metrics/batch"

        request = BatchMetricConfigRequest()
        request.metric_ids = [metric_configuration.id]

        request.metric_configuration = metric_configuration
        request_json = request.to_json()

        response = self._call_datawatch(Method.PUT, url, request_json)

        return BatchMetricConfigResponse().from_dict(response)

    def create_source(self, request: CreateSourceRequest, if_not_exists: bool = False,
                      overwrite: bool = False) -> Source:
        """
        Creates a source in Bigeye.
        :param request: A CreateSourceRequest containing detail for the source to create.
        :param if_not_exists: If true then will not overwrite and return the existing source.  Matches on alias.
        :param overwrite: If true then will overwrite existing source. Matches on aliad.
        :return:
        """

        if if_not_exists and overwrite:
            raise Exception("Cannot create source if not exists and overwrite existing source.. Choose one!")

        if if_not_exists or overwrite:
            existing_sources_ix: Dict[str, Source] = {s.name: s for s in self.get_sources().sources}
            existing = existing_sources_ix.get(request.name, None)
            if existing:
                if if_not_exists:
                    return existing
                if overwrite:
                    self.delete_source(warehouse_id=existing.id)

        url = '/api/v1/sources'

        response = self._call_datawatch(Method.POST, url, request.to_json())
        source = Source().from_dict(value=response['source'])
        log.info(f'Source {source.name} created with warehouse ID: {source.id}')
        return source

    def delete_metric(self, metric: Union[int, MetricConfiguration]):
        """
        Deletes an individual metric.
        Args:
            metric: MetricConfiguration

        Warnings: Will log a warning when attempting to delete a metric created by Bigconfig.
        """

        if isinstance(metric, int):
            metric = self.get_metric_configuration(metric_id=metric)

        if metric.metric_creation_state == MetricCreationState.METRIC_CREATION_STATE_SUITE:
            log.warning(f"Cannot delete metric because it was created by Bigconfig.  "
                        f"MetricID: {metric.id}")
        else:
            url = f'/api/v1/metrics/{metric.id}'

            self._call_datawatch(Method.DELETE, url)

    def delete_source(self, warehouse_id: int):

        url = f'/api/v1/sources/{warehouse_id}'

        self._call_datawatch(Method.DELETE, url)
        log.info(f'Begin delete for warehouse ID: {warehouse_id}')

    def get_issues(
            self,
            *,
            current_status: List[IssueStatus] = [],
            warehouse_ids: List[int] = [],
            schema_ids: List[int] = [],
            metric_ids: List[int] = [],
            collection_ids: List[int] = [],
            issue_ids: List[int] = [],
            schema_names: List[str] = [],
            table_ids: List[int] = [],
            column_ids: List[int] = [],
            related_issue_ids: List[int]= []) -> List[Issue]:

        url = '/api/v1/issues/fetch'

        # TODO: Will require update when GetIssuesRequest refactors to collection_ids

        request = GetIssuesRequest()
        request.current_status = current_status
        request.warehouse_ids = warehouse_ids
        request.schema_ids = schema_ids
        request.metric_ids = metric_ids
        request.sla_ids = collection_ids
        request.issue_ids = issue_ids
        request.schema_names = schema_names
        request.table_ids = table_ids
        request.column_ids = column_ids
        request.related_issue_ids = related_issue_ids

        response = GetIssuesResponse().from_dict(
            self._call_datawatch(method=Method.POST, url=url, body=request.to_json())
        )
        issues: List[Issue] = response.issue
        while response.pagination_info.next_cursor:
            request.page_cursor = response.pagination_info.next_cursor
            response = self._call_datawatch(method=Method.POST, url=url, body=request.to_json())
            issues.extend(
                GetIssuesResponse().from_dict(response).issue
            )
        return issues

    def get_issue(self, issue_id: int) -> Issue:
        return self.get_issues(issue_ids=[issue_id])[0]

    def update_issue(self,
                     *,
                     issue_id: int,
                     issue_status: str,
                     updated_by: str,
                     update_message: str,
                     closing_label: Optional[str]) -> Issue:

        url = f'/api/v1/issues/{issue_id}'

        status = IssueStatus.from_string(f'ISSUE_STATUS_{issue_status.upper()}')
        isu = IssueStatusUpdate()
        isu.updated_by = User(name=updated_by)
        isu.message = update_message
        isu.new_status = status

        if status == IssueStatus.ISSUE_STATUS_CLOSED:
            label = MetricRunLabel.from_string(f'METRIC_RUN_LABEL_{closing_label.upper()}')
            isu.closing_label = label

        request = UpdateIssueRequest(status_update=isu)

        response = self._call_datawatch(Method.PUT, url, request.to_json())
        return UpdateIssueResponse().from_dict(response).issue

    def get_debug_preview(self, *, metric_id: int = 0) -> GetPreviewResponse:
        """Get sample debug rows for debug page"""

        url = f'/api/v1/metrics/{metric_id}/debug/preview'

        return GetPreviewResponse().from_dict(self._call_datawatch(Method.GET, url))

    def create_named_schedule(
            self, *, name: str = "", cron: str = ""
    ) -> NamedSchedule:
        """Create a named schedule"""

        # post: "/api/v1/schedules"

        url = "/api/v1/schedules"

        request = CreateNamedScheduleRequest()
        request.name = name
        request.cron = cron

        r = self._call_datawatch(Method.POST, url, request.to_json())
        return NamedSchedule().from_dict(r)

    def delete_named_schedule(self, *, schedule_id: int = 0) -> Empty:
        """Deleted a named schedule"""

        # delete: "/api/v1/schedules/{schedule_id}"

        url = f"/api/v1/schedules/{schedule_id}"

        return self._call_datawatch(Method.DELETE, url)

    def get_named_schedule(
            self,
            *,
            ids: List[int] = [],
            page_size: int = 0,
            page_cursor: str = "",
            sort_field: NamedScheduleSortField = 0,
            sort_direction: SortDirection = 0,
            search: str = "",
    ) -> GetNamedSchedulesResponse:
        """Get named schedules"""

        # get: "/api/v1/schedules"

        request = GetNamedSchedulesRequest()
        request.ids = ids
        request.page_size = page_size
        request.page_cursor = page_cursor
        request.sort_field = sort_field
        request.sort_direction = sort_direction
        request.search = search

        url = f'/api/v1/schedules{encode_url_params(d=request.to_dict())}'

        response = self._call_datawatch(Method.GET, url=url)

        return GetNamedSchedulesResponse().from_dict(response)

    def create_data_node(self, *, node_type: DataNodeType, node_entity_id: int) -> DataNode:
        """Create Data Node for Lineage"""

        request = CreateDataNodeRequest()
        request.node_type = node_type
        request.node_entity_id = node_entity_id

        url = '/api/v1/lineage/nodes'

        response = self._call_datawatch(Method.POST, url=url, body=request.to_json())

        return DataNode().from_dict(response)

    def get_relationships_for_data_nodes(self, *, data_node_id: int) -> GetLineageRelationshipsForNodeResponse:

        url = f'/api/v1/lineage/nodes/{data_node_id}/relationships'
        return GetLineageRelationshipsForNodeResponse().from_dict(
            self._call_datawatch(Method.GET, url=url)
        )

    def delete_data_node(self, *, data_node_id: int):

        url = f"/api/v1/lineage/nodes/{data_node_id}"
        log.info(f"Deleting data node {data_node_id}")
        self._call_datawatch(Method.DELETE, url=url)

    def create_table_lineage_relationship(self,
                                          *,
                                          upstream_data_node_id: int,
                                          downstream_data_node_id: int) -> LineageRelationship:
        """Create lineage relationship between two tables"""

        request = CreateLineageRelationshipRequest()
        request.upstream_data_node_id = upstream_data_node_id
        request.downstream_data_node_id = downstream_data_node_id

        url = '/api/v1/lineage/relationships'

        response = self._call_datawatch(Method.POST, url=url, body=request.to_json())
        return LineageRelationship().from_dict(response)

    def delete_lineage_relationship(self,
                                    *,
                                    relationship_id: int):

        url = f"/api/v1/lineage/relationships/{relationship_id}"
        log.info(f"Deleting lineage relationship with ID: {relationship_id}")
        self._call_datawatch(Method.DELETE, url=url)

    def delete_lineage_relationship_for_node(self,
                                             *,
                                             data_node_id: int):

        url = f"/api/v1/lineage/nodes/{data_node_id}/relationships"
        log.info(f"Deleting lineage relationships for data node {data_node_id}")
        self._call_datawatch(Method.DELETE, url=url)

    def upsert_metric_template(self,
                               *,
                               id: int,
                               name: str,
                               template: str,
                               return_type: Union[SimpleFieldType.BOOLEAN, SimpleFieldType.NUMERIC],
                               parameters: List[str],
                               source: Source) -> CreateMetricTemplateResponse:

        metric_template = MetricTemplate()
        request = CreateMetricTemplateRequest()
        warehouse = Warehouse(id=source.id, name=source.name, warehouse_vendor=source.database_type)
        params = []

        metric_template.id = id
        metric_template.name = name
        metric_template.template = template
        metric_template.return_type = FieldType.from_string(
            f"{SimpleFieldType.get_protobuf_type_prefix()}_{return_type}")

        for p in parameters:
            kv = p.split("=")
            key = kv[0]
            value = MetricTemplateParameterType.from_string(
                f"{SimpleMetricTemplateParameterType.get_protobuf_type_prefix()}_{kv[1]}"
            )
            temp_param = MetricTemplateParametersFieldEntry(key=key, value=value)
            params.append(temp_param)

        metric_template.parameters = params
        metric_template.source = warehouse

        request.metric_template = metric_template

        url = "/api/v1/metric-templates"

        response = self._call_datawatch(Method.POST, url=url, body=request.to_json())
        return CreateMetricTemplateResponse().from_dict(response)

    def delete_metric_template(self,
                               *,
                               template_id: int):
        url = f"/api/v1/metric-templates/{template_id}"
        log.info(f"Deleting template with ID: {template_id}")
        self._call_datawatch(Method.DELETE, url=url)

    def get_all_metric_templates(self,
                                 *,
                                 page_size: int = 0,
                                 page_cursor: str = "",
                                 sort_field: MetricTemplateSortField = 0,
                                 sort_direction: SortDirection = 0,
                                 search: str = "") -> List[MetricTemplate]:
        request = GetMetricTemplateListRequest()
        request.page_size = page_size
        request.page_cursor = page_cursor
        request.sort_field = sort_field
        request.sort_direction = sort_direction
        request.search = search

        url = "/api/v1/metric-templates/fetch"

        response = GetMetricTemplateListResponse().from_dict(
            self._call_datawatch(method=Method.POST, url=url, body=request.to_json())
        )

        metric_templates: List[MetricTemplate] = response.metric_templates

        while response.pagination_info.next_cursor:
            request.page_cursor = response.pagination_info.next_cursor
            response = GetMetricTemplateListResponse().from_dict(
                self._call_datawatch(method=Method.POST, url=url, body=request.to_json())
            )
            metric_templates.extend(response.metric_templates)

        return metric_templates

    def create_virtual_table(self,
                             *,
                             name: str,
                             sql: str,
                             warehouse_id: int):
        request = VirtualTableRequest(name=name, sql=sql, warehouse_id=warehouse_id)
        url = "/api/v1/virtual-tables"
        self._call_datawatch(Method.POST, url=url, body=request.to_json())

    def update_virtual_table(self,
                             *,
                             id: int,
                             name: str,
                             sql: str,
                             warehouse_id: int):
        update_request = UpdateVirtualTableRequest()
        update_request.id = id
        update_request.request = VirtualTableRequest(name=name, sql=sql, warehouse_id=warehouse_id)

        url = f"/api/v1/virtual-tables/{id}"
        self._call_datawatch(Method.PUT, url=url, body=update_request.to_json())

    def delete_virtual_table(self,
                             *,
                             table_id: int):

        url = f"/api/v1/virtual-tables/{table_id}"
        log.info(f"Deleting virtual table with ID: {table_id}")
        self._call_datawatch(Method.DELETE, url=url)

    def get_all_virtual_tables(self) -> List[VirtualTable]:

        url = "/api/v1/virtual-tables"
        # TODO: Change to response from protobuf when available
        response = [VirtualTable(table=Table().from_dict(vt['table']), sql_query=vt['sqlQuery'])
                    for vt in self._call_datawatch(Method.GET, url=url)]
        return response

    def fetch_schema_changes(self,
                             *,
                             page_size: int = 0,
                             page_cursor: str = "",
                             sort_field: SchemaChangeSortField = 0,
                             sort_direction: SortDirection = 0,
                             search: str = "",
                             source_id: Optional[int] = None,
                             schema_id: Optional[int] = None,
                             table_id: Optional[int] = None,
                             column_id: Optional[int] = None) -> List[SchemaChange]:

        request = GetSchemaChangesRequest()
        request.page_size = page_size
        request.page_cursor = page_cursor
        request.sort_field = sort_field
        request.sort_direction = sort_direction
        request.search = search
        request.source_id = source_id
        request.schema_id = schema_id
        request.table_id = table_id
        request.column_id = column_id

        url = "/api/v1/schema-changes/fetch"

        response = GetSchemaChangesResponse().from_dict(
            self._call_datawatch(Method.POST, url=url, body=request.to_json())
        )

        schema_changes: List[SchemaChange] = response.schema_changes

        while response.pagination_info.next_cursor:
            request.page_cursor = response.pagination_info.next_cursor
            response = GetSchemaChangesResponse().from_dict(
                self._call_datawatch(method=Method.POST, url=url, body=request.to_json())
            )
            schema_changes.extend(response.schema_changes)

        return schema_changes

    def get_integrations(self) -> List[Integration]:

        url = "/api/v1/integrations"
        return IntegrationsResponse().from_dict(
            self._call_datawatch(Method.GET, url=url)
        ).integrations

    def get_integration_entities(self, *, integration_id: int) -> List[TableauWorkbook]:
        # TODO: Will require change once we onboard more integrations.
        url = f"/api/v1/integrations/{integration_id}/entities"
        return GetIntegrationEntitiesResponse().from_dict(
            self._call_datawatch(Method.GET, url=url)
        ).tableau_workbooks

    def get_advanced_configs(self) -> List[ConfigValue]:
        url = "/api/v1/configs/fetch"
        return GetConfigListResponse().from_dict(
            self._call_datawatch(Method.POST, url=url)
        ).advanced_configs

    def rebuild_schema(self, schema_id: int) -> WorkflowResponse:
        url = f"/api/v1/rebuilds/schemas/{schema_id}"
        return WorkflowResponse().from_dict(
            self._call_datawatch(method=Method.POST, url=url)
        )

    def rebuild_source(self, source_id: int) -> WorkflowResponse:
        url = f"/api/v1/rebuilds/sources/{source_id}"
        return WorkflowResponse().from_dict(
            self._call_datawatch(method=Method.POST, url=url)
        )

    def get_workflow_status(self, workflow_id: int) -> WorkflowStatusResponse:
        url = f"/api/v1/workflows/{workflow_id}/status"
        return WorkflowStatusResponse().from_dict(
            self._call_datawatch(method=Method.GET, url=url)
        )

    # def bulk_update_metrics(
    #         self,
    #         *,
    #         where: Optional[WhereClause] = None,
    #         edit_request: Optional[MetricConfiguration] = None,
    #         mute_request: Optional[MuteRequest] = None,
    #         is_run: bool = False,
    #         is_delete: bool = False,
    #         add_to_collection: int = 0,
    #         remove_from_collection: int = 0,
    # ) -> BulkResponse:
    #     """Performs bulk metric updates"""
    #
    #     request = BulkMetricOperation()
    #     if where is not None:
    #         request.where = where
    #     if edit_request is not None:
    #         request.edit_request = edit_request
    #     if mute_request is not None:
    #         request.mute_request = mute_request
    #     request.is_run = is_run
    #     request.is_delete = is_delete
    #     request.add_to_collection = add_to_collection
    #     request.remove_from_collection = remove_from_collection
    #
    #     print(edit_request)
    #
    #     url = f'/api/v1/metrics/bulk'
    #
    #     response = self._call_datawatch(Method.PUT, url=url, params=request.to_dict())
    #
    #     return BulkResponse().from_dict(response)
